package net.f00f.javathrottle;
/*
	Throttle, a class to limit the bandwidth between an input stream
	and an output stream.  Copyright (C) 2003 Ben Damm

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

	For information, contact bdamm@f00f.net
*/

import java.awt.GridLayout;
import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;


import java.io.InputStream;
import java.io.OutputStream;
import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.IOException;

import java.util.Iterator;
import java.util.LinkedList;
import java.util.Observable;
import java.util.Observer;
import java.util.Properties;


/**
 * Throttle acts as a limiter between std in and std out.
 *
 * @author Ben Damm
 * */
public class Throttle extends Thread
{
	private static final int MAX_bufsize = 1024<<5;
	// The maximum (and minimum) time between UI update signals.
	private static final int MAX_emit_update = 500;

	// Maximum desired rate at which throttle is still careful.
	private static final double MAX_careful_rate = 100;
	// Controls how "careful" the program should be with producing
	// the perfect rate, right on time (ie, flush out on every write,
	// etc).
	private boolean hint_careful = true;

	private BufferedInputStream bin;
	private OutputStream bout;
	// bytes/msec
	private Rate desiredRate;
	private Rate actualRate;

	private long total = 0;
	private long windowMaxRead = 0;

	// The window of time within which the rate is calculated
	private long desiredWindow = 10000;
	private long actualWindow = 0;

	private boolean reconfig = false;
	private int desiredBufSize;
	private int actualBufSize;

	private long startTime = 0;
	private long lastUIUpdate = 0;

	private Observable observable;
	private boolean paused = false;

	private boolean shutdown = false;

	// A queue of transactions, only the newest
	private LinkedList dataStamps;


	private static class DataStamp
	{
		public DataStamp( long time, int size )
		{
			this.time = time;
			this.size = size;
		}
		long time;
		int size;
	}

	public Throttle( InputStream in, OutputStream out )
	{
		bin = new BufferedInputStream( in );
		bout = out;

		// Set default to no data
		desiredRate = new Rate();
		desiredBufSize = 0; // bytes (0 is automatic)
		actualBufSize = 1; // 
		actualRate  = new Rate();

		dataStamps = new LinkedList();

		observable = new Observable()
		{
			public void notifyObservers()
			{
				setChanged();
				super.notifyObservers();
			}
		};
	}

	public void togglePaused()
	{
		synchronized( this )
		{
			paused = !paused;
			notifyAll();
		}
	}

	public void setPaused( boolean bool )
	{
		synchronized( this )
		{
			paused = bool;
			notifyAll();
		}
	}

	/**
	 * Ensure data in the throttle is flushed.
	 * */
	public void flush()
	{
	}

	public void addObserver( Observer obs )
	{
		observable.addObserver( obs );
	}

	public long getTotal()
	{
		return total;
	}

	public Rate getActualRate()
	{
		return actualRate;
	}

	public int getActualBufSize()
	{
		return actualBufSize;
	}

	public long getMaxWindowRead()
	{
		return windowMaxRead;
	}

	public long getActualWindow()
	{
		return actualWindow;
	}

	public long getDesiredWindow()
	{
		return desiredWindow;
	}

	public long getStartTime()
	{
		return startTime;
	}

	private void fireUpdate()
	{
		observable.notifyObservers();
	}

	public void setDesiredRate( Rate sdr )
	{
		if( sdr == null )
			desiredRate = Unit.Bps.makeRate( 1 );
		else 
			desiredRate = sdr;

		actualRate.setUnit( desiredRate.getUnit() );

		hint_careful = desiredRate.getBytesPerMSec() < MAX_careful_rate;

		// reset the window
		actualWindow = 0;

		// If the buffer is automatic, resize it
		if( desiredBufSize == 0 )
		{
			// desiredRate be infinity
			actualBufSize = Math.min( MAX_bufsize, (int)(desiredRate.getBytesPerMSec()*100) );
			actualBufSize = Math.max( 1, actualBufSize );

			log( "Set actualBufSize: " + actualBufSize );
		}
		synchronized( this )
		{
			reconfig = true;
		}
	}

	private void log(String msg)
	{
		Main.debug( msg );
	}

	public Rate getDesiredRate()
	{
		return desiredRate;
	}

	public synchronized void shutdown()
	{
		shutdown = true;
		setPaused( false );
	}

	public void run()
	{
		try
		{
			synchronized( this )
			{
				int bSize = actualBufSize;
				log( "Setting bSize to " + bSize );

				bin = new BufferedInputStream( System.in );
				bout = new BufferedOutputStream( System.out );

				reconfig = true;
			}

			while( true )
			{
				synchronized( this )
				{
					if( !reconfig )
						break;
					if( shutdown )
						break;
					reconfig = false;
				}
				int dataLen = actualBufSize;
				byte data[] = new byte[dataLen];
				boolean stallAgain = false;

				//log("Beep");
				//
				// Main loop
				while( true )
				{
					pipeData(data, dataLen);

					if( shutdown )
					{
						bout.flush();
						break;
					}

					if( reconfig ) break;

					stall();
				}
			}
		}
		catch( IOException ioe )
		{
			log( "Error: " + ioe );
			ioe.printStackTrace(System.err);
		}
	}

	/**
	 * Use the given buffer to pipe data from one stream to the other.  
	 *
	 * */
	private synchronized void pipeData(byte[] buffer, int dataLen)
		throws IOException
	{
		int doRead = Math.min( dataLen, bin.available());

		// Apparently. if we are at end of stream,
		// bin.available will return 0.  But if we ask
		// bin.read to read 0 bytes, it'll return zero
		// bytes instead of -1, for end of stream.
		//
		// Thus, we must try to read at least one byte.
		if( doRead == 0 ) doRead = 1;

		log("boop: " + " doRead: " + doRead + " dataLen: " + dataLen );
		int readSize = bin.read( buffer, 0, doRead );
		if( readSize == -1 )
			streamOver();

		if( startTime == 0 )
			startTime = System.currentTimeMillis();

		bout.write( buffer, 0, readSize );
		if( hint_careful )
		{
			bout.flush();
		}

		long now = System.currentTimeMillis();
		//synchronized( dataStamps )
		{
			dataStamps.add( new DataStamp(now, readSize) );
		}

		total += readSize;
	}

	/**
	 * @return false if we are ready to read data, true to sleep
	 * again.
	 * */
	private void stall()
	{
		synchronized ( this )
		{
			while( paused )
			{
				log(" --------- paused!" );
				try
				{
					wait();
				}
				catch( InterruptedException ie )
				{
					throw new IllegalStateException("Interrupted while paused, Stopping!");
				}
			}
		}

		maybeEmitUpdate();

		// calc rate
		// if rate is lower than desired, return
		// if rate is higher than desired, wait a bit and chech again

		int wakeUpEarly = 0; // Wake up xx ms earlier
		double rateDiff = calcRate() - desiredRate.getBytesPerMSec();
		double thresh = 0;
		while( rateDiff > thresh )
		{
			//int sleepTime = (int)(1/actualRate) - wakeUpEarly;
			//sleepTime = Math.max(1,sleepTime);
			int sleepTime = 200;

			log( "ratediff: " + rateDiff + " -- " + thresh + " -- " + sleepTime);
			try
			{
				Thread.sleep(sleepTime);
			}
			catch( InterruptedException ie )
			{
				throw new IllegalStateException("Rate reduction thread interrupted? Stopping!");
			}

			maybeEmitUpdate();
			rateDiff = calcRate() - desiredRate.getBytesPerMSec();
		}
	}

	public void maybeEmitUpdate()
	{
		long now = System.currentTimeMillis();
		if( now - lastUIUpdate > MAX_emit_update )
		{
			lastUIUpdate = now;
			fireUpdate();
		}
	}

	/**
	 * calculates the current rate, returning the result.  Side
	 * effect, sets the actualRate variable.
	 * */
	private double calcRate()
	{
		long now = System.currentTimeMillis();
		long ago = now - actualWindow;
		long first = Math.max( ago, startTime);

		// If the window is smaller than it should be, increase it a
		// bit (1% of the difference in size)
		if( actualWindow < desiredWindow )
			actualWindow += (long)((desiredWindow-actualWindow)/100.0);

		long windowTotal = 0;

		int dsSz1 = dataStamps.size();
		//synchronized( dataStamps )
		{
			Iterator i = dataStamps.iterator();
			while( i.hasNext())
			{
				DataStamp ds = (DataStamp)i.next();
				if( ds.time < ago ) 
				{
					i.remove();
					if( ds.size == windowMaxRead )
						windowMaxRead = 0;
				}
				else
				{
					long sz = ds.size;
					windowTotal += sz;
					if( sz > windowMaxRead )
						windowMaxRead = sz;
				}
			}
		}
		int dsSz2 = dataStamps.size();
		//log( " Now: " + now + " Ago: " + ago + " First: " + first );
		//log( "DS Size1: " + dsSz1 + " DS Size2: " + dsSz2 + " wT: " + windowTotal);

		double result = (double)windowTotal / (now - first);
		actualRate.setBytesPerMSec( result );
		//log( "Rate: " + actualRate );
		//log( "Target: " + desiredRate );
		return result;
	}

	private void streamOver()
		throws IOException
	{
		log( "done" );
		bout.flush();  // redundant?
		bout.close();
		System.exit(0);
	}



}

